@@ -1,4 +1,4 @@
-# Copyright (C) 2011-2017 2ndQuadrant Limited
+# Copyright (C) 2011-2020 2ndQuadrant Limited
 #
 # This file is part of Barman.
 #
@@ -19,11 +19,13 @@
 This module implements the interface with the command line and the logger.
 """
 
+import json
 import logging
 import os
 import sys
 from argparse import SUPPRESS, ArgumentTypeError
 from contextlib import closing
+from functools import wraps
 
 from argh import ArghParser, arg, expects_obj, named
 
@@ -31,50 +33,30 @@ import barman.config
 import barman.diagnose
 from barman import output
 from barman.config import RecoveryOptions
-from barman.exceptions import BadXlogSegmentName
+from barman.exceptions import BadXlogSegmentName, RecoveryException, SyncError
 from barman.infofile import BackupInfo
 from barman.server import Server
-from barman.utils import configure_logging, drop_privileges, parse_log_level
+from barman.utils import (BarmanEncoder, check_non_negative, check_positive,
+                          configure_logging, drop_privileges, force_str,
+                          get_log_levels, parse_log_level)
 
 _logger = logging.getLogger(__name__)
 
 
-def check_non_negative(value):
+def check_target_action(value):
     """
-    Check for a positive integer option
+    Check the target action option
 
     :param value: str containing the value to check
     """
     if value is None:
         return None
-    try:
-        int_value = int(value)
-    except Exception:
-        raise ArgumentTypeError("'%s' is not a valid non negative integer" %
-                                value)
-    if int_value < 0:
-        raise ArgumentTypeError("'%s' is not a valid non negative integer" %
-                                value)
-    return int_value
 
+    if value in ('pause', 'shutdown', 'promote'):
+        return value
 
-def check_positive(value):
-    """
-    Check for a positive integer option
-
-    :param value: str containing the value to check
-    """
-    if value is None:
-        return None
-    try:
-        int_value = int(value)
-    except Exception:
-        raise ArgumentTypeError("'%s' is not a valid positive integer" %
-                                value)
-    if int_value < 1:
-        raise ArgumentTypeError("'%s' is not a valid positive integer" %
-                                value)
-    return int_value
+    raise ArgumentTypeError("'%s' is not a valid recovery target action" %
+                            value)
 
 
 @named('list-server')
@@ -94,18 +76,24 @@ def list_server(minimal=False):
         # we want all servers and the server cannot be None
 
         output.init('list_server', name, minimal=minimal)
-        description = server.config.description
+        description = server.config.description or ''
         # If the server has been manually disabled
         if not server.config.active:
             description += " (inactive)"
         # If server has configuration errors
         elif server.config.disabled:
             description += " (WARNING: disabled)"
+        # If server is a passive node
+        if server.passive_node:
+            description += ' (Passive)'
         output.result('list_server', name, description)
     output.close_and_exit()
 
 
-def cron():
+@arg('--keep-descriptors',
+     help='Keep the stdout and the stderr streams attached '
+          'to Barman subprocesses.')
+def cron(keep_descriptors=False):
     """
     Run maintenance tasks (global command)
     """
@@ -119,7 +107,16 @@ def cron():
         # server is None and to report inactive and disabled servers,
         # but here we have only active and well configured servers.
 
-        server.cron()
+        try:
+            server.cron(keep_descriptors=keep_descriptors)
+        except Exception:
+            # A cron should never raise an exception, so this code
+            # should never be executed. However, it is here to protect
+            # unrelated servers in case of unexpected failures.
+            output.exception(
+                "Unable to run cron on server '%s', "
+                "please look in the barman log file for more details.",
+                name)
 
     output.close_and_exit()
 
@@ -167,7 +164,7 @@ def backup_completer(prefix, parsed_args, **kwargs):
      action='store_true',
      default=SUPPRESS)
 @arg('--no-immediate-checkpoint',
-     help='forces the initial checkpoint to be spreaded',
+     help='forces the initial checkpoint to be spread',
      dest='immediate_checkpoint',
      action='store_false',
      default=SUPPRESS)
@@ -187,12 +184,31 @@ def backup_completer(prefix, parsed_args, **kwargs):
 @arg('--jobs', '-j',
      help='Run the copy in parallel using NJOBS processes.',
      type=check_positive, metavar='NJOBS')
+@arg('--bwlimit',
+     help="maximum transfer rate in kilobytes per second. "
+          "A value of 0 means no limit. Overrides 'bandwidth_limit' "
+          "configuration option.",
+     metavar='KBPS',
+     type=check_non_negative,
+     default=SUPPRESS)
+@arg('--wait', '-w',
+     help='wait for all the required WAL files to be archived',
+     dest='wait',
+     action='store_true',
+     default=False)
+@arg('--wait-timeout',
+     help='the time, in seconds, spent waiting for the required '
+          'WAL files to be archived before timing out',
+     dest='wait_timeout',
+     metavar='TIMEOUT',
+     default=None,
+     type=check_non_negative)
 @expects_obj
 def backup(args):
     """
     Perform a full backup for the given server (supports 'all')
     """
-    servers = get_server_list(args, skip_inactive=True)
+    servers = get_server_list(args, skip_inactive=True, skip_passive=True)
     for name in sorted(servers):
         server = servers[name]
 
@@ -210,8 +226,10 @@ def backup(args):
             server.config.immediate_checkpoint = args.immediate_checkpoint
         if args.jobs is not None:
             server.config.parallel_jobs = args.jobs
+        if hasattr(args, 'bwlimit'):
+            server.config.bandwidth_limit = args.bwlimit
         with closing(server):
-            server.backup()
+            server.backup(wait=args.wait, wait_timeout=args.wait_timeout)
     output.close_and_exit()
 
 
@@ -323,11 +341,16 @@ def rebuild_xlogdb(args):
      help='target time. You can use any valid unambiguous representation. '
           'e.g: "YYYY-MM-DD HH:MM:SS.mmm"')
 @arg('--target-xid', help='target transaction ID')
+@arg('--target-lsn', help='target LSN (Log Sequence Number)')
 @arg('--target-name',
      help='target name created previously with '
           'pg_create_restore_point() function call')
+@arg('--target-immediate',
+     help='end recovery as soon as a consistent state is reached',
+     action='store_true',
+     default=False)
 @arg('--exclusive',
-     help='set target xid to be non inclusive', action="store_true")
+     help='set target to be non inclusive', action="store_true")
 @arg('--tablespace',
      help='tablespace relocation rule',
      metavar='NAME:LOCATION', action='append')
@@ -343,6 +366,13 @@ def rebuild_xlogdb(args):
      help='specifies the backup ID to recover')
 @arg('destination_directory',
      help='the directory where the new server is created')
+@arg('--bwlimit',
+     help="maximum transfer rate in kilobytes per second. "
+          "A value of 0 means no limit. Overrides 'bandwidth_limit' "
+          "configuration option.",
+     metavar='KBPS',
+     type=check_non_negative,
+     default=SUPPRESS)
 @arg('--retry-times',
      help='Number of retries after an error if base backup copy fails.',
      type=check_non_negative)
@@ -364,16 +394,41 @@ def rebuild_xlogdb(args):
      dest='get_wal',
      action='store_false',
      default=SUPPRESS)
+@arg('--network-compression',
+     help='Enable network compression during remote recovery.',
+     dest='network_compression',
+     action='store_true',
+     default=SUPPRESS)
+@arg('--no-network-compression',
+     help='Disable network compression during remote recovery.',
+     dest='network_compression',
+     action='store_false',
+     default=SUPPRESS)
+@arg('--target-action',
+     help='Specifies what action the server should take once the '
+          'recovery target is reached. This option is not allowed for '
+          'PostgreSQL < 9.1. If PostgreSQL is between 9.1 and 9.4 included '
+          'the only allowed value is "pause". If PostgreSQL is 9.5 or newer '
+          'the possible values are "shutdown", "pause", "promote".',
+     dest='target_action',
+     type=check_target_action,
+     default=SUPPRESS)
+@arg('--standby-mode',
+     dest="standby_mode",
+     action='store_true',
+     default=SUPPRESS,
+     help='Enable standby mode when starting '
+          'the recovered PostgreSQL instance')
 @expects_obj
 def recover(args):
     """
-    Recover a server at a given time or xid
+    Recover a server at a given time, name, LSN or xid
     """
     server = get_server(args)
 
     # Retrieves the backup
     backup_id = parse_backup_id(server, args)
-    if backup_id.status != BackupInfo.DONE:
+    if backup_id.status not in BackupInfo.STATUS_COPY_DONE:
         output.error(
             "Cannot recover from backup '%s' of server '%s': "
             "backup status is not DONE",
@@ -425,16 +480,55 @@ def recover(args):
             server.config.recovery_options.remove(RecoveryOptions.GET_WAL)
     if args.jobs is not None:
         server.config.parallel_jobs = args.jobs
+    if hasattr(args, 'bwlimit'):
+        server.config.bandwidth_limit = args.bwlimit
+
+    # PostgreSQL supports multiple parameters to specify when the recovery
+    # process will end, and in that case the last entry in recovery
+    # configuration files will be used. See [1]
+    #
+    # Since the meaning of the target options is not dependent on the order
+    # of parameters, we decided to make the target options mutually exclusive.
+    #
+    # [1]: https://www.postgresql.org/docs/current/static/
+    #   recovery-target-settings.html
+
+    target_options = ['target_tli', 'target_time', 'target_xid',
+                      'target_lsn', 'target_name', 'target_immediate']
+    specified_target_options = len(
+        [option for option in target_options if getattr(args, option)])
+    if specified_target_options > 1:
+        output.error(
+            "You cannot specify multiple targets for the recovery operation")
+        output.close_and_exit()
+
+    if hasattr(args, 'network_compression'):
+        if args.network_compression and args.remote_ssh_command is None:
+            output.error(
+                "Network compression can only be used with "
+                "remote recovery.\n"
+                "HINT: If you want to do a remote recovery "
+                "you have to use the --remote-ssh-command option")
+            output.close_and_exit()
+        server.config.network_compression = args.network_compression
+
     with closing(server):
-        server.recover(backup_id,
-                       args.destination_directory,
-                       tablespaces=tablespaces,
-                       target_tli=args.target_tli,
-                       target_time=args.target_time,
-                       target_xid=args.target_xid,
-                       target_name=args.target_name,
-                       exclusive=args.exclusive,
-                       remote_command=args.remote_ssh_command)
+        try:
+            server.recover(backup_id,
+                           args.destination_directory,
+                           tablespaces=tablespaces,
+                           target_tli=args.target_tli,
+                           target_time=args.target_time,
+                           target_xid=args.target_xid,
+                           target_lsn=args.target_lsn,
+                           target_name=args.target_name,
+                           target_immediate=args.target_immediate,
+                           exclusive=args.exclusive,
+                           remote_command=args.remote_ssh_command,
+                           target_action=getattr(args, 'target_action', None),
+                           standby_mode=getattr(args, 'standby_mode', None))
+        except RecoveryException as exc:
+            output.error(force_str(exc))
 
     output.close_and_exit()
 
@@ -471,30 +565,30 @@ def show_server(args):
     output.close_and_exit()
 
 
-@named('switch-xlog')
+@named('switch-wal')
 @arg('server_name', nargs='+',
      completer=server_completer_all,
-     help="specifies the server name target of the switch-xlog command")
+     help="specifies the server name target of the switch-wal command")
 @arg('--force',
-     help='forces the switch of a xlog by executing a checkpoint before',
+     help='forces the switch of a WAL by executing a checkpoint before',
      dest='force',
      action='store_true',
      default=False)
 @arg('--archive',
-     help='wait for one xlog file to be archived',
+     help='wait for one WAL file to be archived',
      dest='archive',
      action='store_true',
      default=False)
 @arg('--archive-timeout',
-     help='the time, in seconds, the archiver will wait for a new xlog file '
+     help='the time, in seconds, the archiver will wait for a new WAL file '
           'to be archived before timing out',
      metavar='TIMEOUT',
      default='30',
      type=check_non_negative)
 @expects_obj
-def switch_xlog(args):
+def switch_wal(args):
     """
-    Execute the switch-xlog command on the target server
+    Execute the switch-wal command on the target server
     """
     servers = get_server_list(args, skip_inactive=True)
     for name in sorted(servers):
@@ -503,10 +597,20 @@ def switch_xlog(args):
         if not manage_server_command(server, name):
             continue
         with closing(server):
-            server.switch_xlog(args.force, args.archive, args.archive_timeout)
+            server.switch_wal(args.force, args.archive, args.archive_timeout)
     output.close_and_exit()
 
 
+@named('switch-xlog')
+# Set switch-xlog as alias of switch-wal.
+# We cannot use the @argh.aliases decorator, because it needs Python >= 3.2,
+# so we create a wraqpper function and use @wraps to copy all the function
+# attributes needed by argh
+@wraps(switch_wal)
+def switch_xlog(args):
+    return switch_wal(args)
+
+
 @arg('server_name', nargs='+',
      completer=server_completer_all,
      help="specifies the server names to check "
@@ -532,13 +636,8 @@ def check(args):
                 skip_disabled=False, disabled_is_error=False):
             continue
 
-        # If the server has been manually disabled
-        if not server.config.active:
-            name += " (inactive)"
-        # If server has configuration errors
-        elif server.config.disabled:
-            name += " (WARNING: disabled)"
-        output.init('check', name, server.config.active)
+        output.init('check', name,
+                    server.config.active, server.config.disabled)
         with closing(server):
             server.check()
     output.close_and_exit()
@@ -556,6 +655,83 @@ def diagnose():
     output.close_and_exit()
 
 
+@named('sync-info')
+@arg('--primary', help='execute the sync-info on the primary node (if set)',
+     action='store_true', default=SUPPRESS)
+@arg("server_name",
+     completer=server_completer,
+     help='specifies the server name for the command')
+@arg("last_wal",
+     help='specifies the name of the latest WAL read',
+     nargs='?')
+@arg("last_position",
+     nargs='?',
+     type=check_positive,
+     help='the last position read from xlog database (in bytes)')
+@expects_obj
+def sync_info(args):
+    """
+    Output the internal synchronisation status.
+    Used to sync_backup with a passive node
+    """
+    server = get_server(args)
+    try:
+        # if called with --primary option
+        if getattr(args, 'primary', False):
+            primary_info = server.primary_node_info(args.last_wal,
+                                                    args.last_position)
+            output.info(json.dumps(primary_info, cls=BarmanEncoder, indent=4),
+                        log=False)
+        else:
+            server.sync_status(args.last_wal, args.last_position)
+    except SyncError as e:
+        # Catch SyncError exceptions and output only the error message,
+        # preventing from logging the stack trace
+        output.error(e)
+
+    output.close_and_exit()
+
+
+@named('sync-backup')
+@arg("server_name",
+     completer=server_completer,
+     help='specifies the server name for the command')
+@arg("backup_id",
+     help='specifies the backup ID to be copied on the passive node')
+@expects_obj
+def sync_backup(args):
+    """
+    Command that synchronises a backup from a master to a passive node
+    """
+    server = get_server(args)
+    try:
+        server.sync_backup(args.backup_id)
+    except SyncError as e:
+        # Catch SyncError exceptions and output only the error message,
+        # preventing from logging the stack trace
+        output.error(e)
+    output.close_and_exit()
+
+
+@named('sync-wals')
+@arg("server_name",
+     completer=server_completer,
+     help='specifies the server name for the command')
+@expects_obj
+def sync_wals(args):
+    """
+    Command that synchronises WAL files from a master to a passive node
+    """
+    server = get_server(args)
+    try:
+        server.sync_wals()
+    except SyncError as e:
+        # Catch SyncError exceptions and output only the error message,
+        # preventing from logging the stack trace
+        output.error(e)
+    output.close_and_exit()
+
+
 @named('show-backup')
 @arg('server_name',
      completer=server_completer,
@@ -609,7 +785,7 @@ def list_files(args):
             "invalid xlog segment name %r\n"
             "HINT: Please run \"barman rebuild-xlogdb %s\" "
             "to solve this issue",
-            str(e), server.config.name)
+            force_str(e), server.config.name)
         output.close_and_exit()
 
 
@@ -629,7 +805,9 @@ def delete(args):
     # Retrieves the backup
     backup_id = parse_backup_id(server, args)
     with closing(server):
-        server.delete_backup(backup_id)
+        if not server.delete_backup(backup_id):
+            output.error("Cannot delete backup (%s %s)"
+                         % (server.config.name, backup_id))
     output.close_and_exit()
 
 
@@ -643,6 +821,9 @@ def delete(args):
      help='put the retrieved WAL file in this directory '
           'with the original name',
      default=SUPPRESS)
+@arg('--partial', '-P',
+     help='retrieve also partial WAL files (.partial)',
+     action='store_true', dest='partial', default=False)
 @arg('--gzip', '-z', '-x',
      help='compress the output with gzip',
      action='store_const', const='gzip', dest='compression', default=SUPPRESS)
@@ -657,6 +838,12 @@ def delete(args):
      metavar='SIZE',
      type=check_positive,
      default=SUPPRESS)
+@arg('--test', '-t',
+     help="test both the connection and the configuration of the requested "
+          "PostgreSQL server in Barman for WAL retrieval. With this option, "
+          "the 'wal_name' mandatory argument is ignored.",
+     action='store_true',
+     default=SUPPRESS)
 @expects_obj
 def get_wal(args):
     """
@@ -666,6 +853,11 @@ def get_wal(args):
     """
     server = get_server(args, inactive_is_error=True)
 
+    if getattr(args, 'test', None):
+        output.info("Ready to retrieve WAL files from the server %s",
+                    server.config.name)
+        return
+
     # Retrieve optional arguments. If an argument is not specified,
     # the namespace doesn't contain it due to SUPPRESS default.
     # In that case we pick 'None' using getattr third argument.
@@ -677,7 +869,42 @@ def get_wal(args):
         server.get_wal(args.wal_name,
                        compression=compression,
                        output_directory=output_directory,
-                       peek=peek)
+                       peek=peek,
+                       partial=args.partial)
+    output.close_and_exit()
+
+
+@named('put-wal')
+@arg('server_name',
+     completer=server_completer,
+     help='specifies the server name for the command')
+@arg('--test', '-t',
+     help='test both the connection and the configuration of the requested '
+          'PostgreSQL server in Barman to make sure it is ready to receive '
+          'WAL files.',
+     action='store_true',
+     default=SUPPRESS)
+@expects_obj
+def put_wal(args):
+    """
+    Receive a WAL file from SERVER_NAME and securely store it in the incoming
+    directory. The file will be read from standard input in tar format.
+    """
+    server = get_server(args, inactive_is_error=True)
+
+    if getattr(args, 'test', None):
+        output.info("Ready to accept WAL files for the server %s",
+                    server.config.name)
+        return
+
+    try:
+        # Python 3.x
+        stream = sys.stdin.buffer
+    except AttributeError:
+        # Python 2.x
+        stream = sys.stdin
+    with closing(server):
+        server.put_wal(stream)
     output.close_and_exit()
 
 
@@ -738,6 +965,32 @@ def receive_wal(args):
     output.close_and_exit()
 
 
+@named('check-backup')
+@arg('server_name',
+     completer=server_completer,
+     help='specifies the server name for the command')
+@arg('backup_id',
+     completer=backup_completer,
+     help='specifies the backup ID')
+@expects_obj
+def check_backup(args):
+    """
+    Make sure that all the required WAL files to check
+    the consistency of a physical backup (that is, from the
+    beginning to the end of the full backup) are correctly
+    archived. This command is automatically invoked by the
+    cron command and at the end of every backup operation.
+    """
+    server = get_server(args)
+
+    # Retrieves the backup
+    backup_info = parse_backup_id(server, args)
+
+    with closing(server):
+        server.check_backup(backup_info)
+    output.close_and_exit()
+
+
 def pretty_args(args):
     """
     Prettify the given argh namespace to be human readable
@@ -782,19 +1035,30 @@ def global_config(args):
         raise SystemExit(msg)
 
     # configure logging
+    if hasattr(args, 'log_level'):
+        config.log_level = args.log_level
     log_level = parse_log_level(config.log_level)
     configure_logging(config.log_file,
                       log_level or barman.config.DEFAULT_LOG_LEVEL,
                       config.log_format)
     if log_level is None:
-        _logger.warn('unknown log_level in config file: %s', config.log_level)
+        _logger.warning('unknown log_level in config file: %s',
+                        config.log_level)
 
-    # configure output
+    # Configure output
     if args.format != output.DEFAULT_WRITER or args.quiet or args.debug:
         output.set_output_writer(args.format,
                                  quiet=args.quiet,
                                  debug=args.debug)
 
+    # Configure color output
+    if args.color == 'auto':
+        # Enable colored output if both stdout and stderr are TTYs
+        output.ansi_colors_enabled = (
+            sys.stdout.isatty() and sys.stderr.isatty())
+    else:
+        output.ansi_colors_enabled = args.color == 'always'
+
     # Load additional configuration files
     config.load_configuration_files_directory()
     # We must validate the configuration here in order to have
@@ -806,6 +1070,7 @@ def global_config(args):
 
 
 def get_server(args, skip_inactive=True, skip_disabled=False,
+               skip_passive=False,
                inactive_is_error=False,
                on_error_stop=True, suppress_error=False):
     """
@@ -821,10 +1086,11 @@ def get_server(args, skip_inactive=True, skip_disabled=False,
         WARNING: the function modifies the content of this parameter
     :param bool skip_inactive: do nothing if the server is inactive
     :param bool skip_disabled: do nothing if the server is disabled
+    :param bool skip_passive: do nothing if the server is passive
     :param bool inactive_is_error: treat inactive server as error
     :param bool on_error_stop: stop if an error is found
     :param bool suppress_error: suppress display of errors (e.g. diagnose)
-    :rtype: barman.server.Server|None
+    :rtype: Server|None
     """
     # This function must to be called with in a single-server context
     name = args.server_name
@@ -847,6 +1113,7 @@ def get_server(args, skip_inactive=True, skip_disabled=False,
 
     # Retrieve the requested server
     servers = get_server_list(args, skip_inactive, skip_disabled,
+                              skip_passive,
                               on_error_stop, suppress_error)
 
     # The requested server has been excluded from get_server_list result
@@ -875,6 +1142,7 @@ def get_server(args, skip_inactive=True, skip_disabled=False,
 
 
 def get_server_list(args=None, skip_inactive=False, skip_disabled=False,
+                    skip_passive=False,
                     on_error_stop=True, suppress_error=False):
     """
     Get the server list from the configuration
@@ -885,9 +1153,10 @@ def get_server_list(args=None, skip_inactive=False, skip_disabled=False,
     :param args: an argparse namespace containing a list server_name parameter
     :param bool skip_inactive: skip inactive servers when 'all' is required
     :param bool skip_disabled: skip disabled servers when 'all' is required
+    :param bool skip_passive: skip passive servers when 'all' is required
     :param bool on_error_stop: stop if an error is found
     :param bool suppress_error: suppress display of errors (e.g. diagnose)
-    :rtype: dict(str,barman.server.Server|None)
+    :rtype: dict[str,Server]
     """
     server_dict = {}
 
@@ -923,7 +1192,8 @@ def get_server_list(args=None, skip_inactive=False, skip_disabled=False,
             output.error("You cannot use 'all' with other server names")
         servers = available_servers
     else:
-        servers = args.server_name
+        # Put servers in a set, so multiple occurrences are counted only once
+        servers = set(args.server_name)
 
     # Loop through all the requested servers
     for server in servers:
@@ -943,6 +1213,11 @@ def get_server_list(args=None, skip_inactive=False, skip_disabled=False,
                 output.info("Skipping temporarily disabled server '%s'"
                             % conf.name)
                 continue
+            # Skip passive nodes, if requested
+            if skip_passive and server_object.passive_node:
+                output.info("Skipping passive server '%s'",
+                            conf.name)
+                continue
             server_dict[server] = server_object
 
     return server_dict
@@ -1005,7 +1280,7 @@ def parse_backup_id(server, args):
 
     :param Server server: server object to search for the required backup
     :param args: command lien arguments namespace
-    :rtype: BackupInfo
+    :rtype: barman.infofile.LocalBackupInfo
     """
     if args.backup_id in ('latest', 'last'):
         backup_id = server.get_last_backup_id()
@@ -1035,6 +1310,14 @@ def main():
                         '(defaults: %s)'
                         % ', '.join(barman.config.Config.CONFIG_FILES),
                    default=SUPPRESS)
+    p.add_argument('--color', '--colour',
+                   help='Whether to use colors in the output',
+                   choices=['never', 'always', 'auto'],
+                   default='auto')
+    p.add_argument('--log-level',
+                   help='Override the default log level',
+                   choices=list(get_log_levels()),
+                   default=SUPPRESS)
     p.add_argument('-q', '--quiet', help='be quiet', action='store_true')
     p.add_argument('-d', '--debug', help='debug output', action='store_true')
     p.add_argument('-f', '--format', help='output format',
@@ -1045,6 +1328,7 @@ def main():
             archive_wal,
             backup,
             check,
+            check_backup,
             cron,
             delete,
             diagnose,
@@ -1052,6 +1336,7 @@ def main():
             list_backup,
             list_files,
             list_server,
+            put_wal,
             rebuild_xlogdb,
             receive_wal,
             recover,
@@ -1059,7 +1344,11 @@ def main():
             show_server,
             replication_status,
             status,
+            switch_wal,
             switch_xlog,
+            sync_info,
+            sync_backup,
+            sync_wals,
         ]
     )
     # noinspection PyBroadException
