@@ -1,4 +1,4 @@
-# Copyright (C) 2011-2017 2ndQuadrant Limited
+# Copyright (C) 2011-2020 2ndQuadrant Limited
 #
 # This file is part of Barman.
 #
@@ -25,6 +25,7 @@ import errno
 import inspect
 import logging
 import os
+import re
 import select
 import signal
 import subprocess
@@ -79,7 +80,7 @@ class StreamLineProcessor(object):
             # Handle the last line (always incomplete, maybe empty)
             self._handler(self._buf)
             return True
-        self._buf += data.decode('utf-8')
+        self._buf += data.decode('utf-8', 'replace')
         # If no '\n' is present, we just read a part of a very long line.
         # Nothing to do at the moment.
         if '\n' not in self._buf:
@@ -131,7 +132,7 @@ class Command(object):
         a maximum of `retry_times` times, waiting for `retry_sleep` seconds
         between every attempt.
 
-        Everytime a command is retried the `retry_handler` is executed
+        Every time a command is retried the `retry_handler` is executed
         before running the command again. The retry_handler must be a callable
         that accepts the following fields:
 
@@ -298,7 +299,7 @@ class Command(object):
                         # there is still an error, exit raising
                         # a CommandMaxRetryExceeded exception and wrap the
                         # original one
-                        raise CommandMaxRetryExceeded(exc)
+                        raise CommandMaxRetryExceeded(*exc.args)
 
     def _get_output_once(self, *args, **kwargs):
         """
@@ -334,6 +335,7 @@ class Command(object):
                      check=False, *args, **kwargs)
         self.out = '\n'.join(out)
         self.err = '\n'.join(err)
+
         _logger.debug("Command stdout: %s", self.out)
         _logger.debug("Command stderr: %s", self.err)
 
@@ -442,7 +444,7 @@ class Command(object):
             will be closed before the child process is executed.
         :rtype: subprocess.Popen
         """
-        # Append the argument provided to this method ot the base argument list
+        # Append the argument provided to this method of the base argument list
         args = self.args + list(args)
         # If shell is True, properly quote the command
         if self.shell:
@@ -573,7 +575,8 @@ class Rsync(Command):
                  include=None, network_compression=None, path=None, **kwargs):
         """
         :param str rsync: rsync executable name
-        :param list[str]|None args: List of additional argument to aways append
+        :param list[str]|None args: List of additional argument to always
+            append
         :param str ssh: the ssh executable to be used when building
             the `-e` argument
         :param list[str] ssh_options: the ssh options to be used when building
@@ -679,7 +682,16 @@ class PostgreSQLClient(Command):
     Superclass of all the PostgreSQL client commands.
     """
 
-    COMMAND = None
+    COMMAND_ALTERNATIVES = None
+    """
+    Sometimes the name of a command has been changed during the PostgreSQL
+    evolution. I.e. that happened with pg_receivexlog, that has been renamed
+    to pg_receivewal. In that case, we should try using pg_receivewal (the
+    newer auternative) and, if that command doesn't exist, we should try
+    using `pg_receivewal`.
+
+    This is a list of command names to be used to find the installed command.
+    """
 
     def __init__(self,
                  connection,
@@ -718,6 +730,74 @@ class PostgreSQLClient(Command):
         self.enable_signal_forwarding(signal.SIGINT)
         self.enable_signal_forwarding(signal.SIGTERM)
 
+    @classmethod
+    def find_command(cls, path=None):
+        """
+        Find the active command, given all the alternatives as set in the
+        property named `COMMAND_ALTERNATIVES` in this class.
+
+        :param str path: The path to use while searching for the command
+        :rtype: Command
+        """
+
+        # TODO: Unit tests of this one
+
+        # To search for an available command, testing if the command
+        # exists in PATH is not sufficient. Debian will install wrappers for
+        # all commands, even if the real command doesn't work.
+        #
+        # I.e. we may have a wrapper for `pg_receivewal` even it PostgreSQL
+        # 10 isn't installed.
+        #
+        # This is an example of what can happen in this case:
+        #
+        # ```
+        # $ pg_receivewal --version; echo $?
+        # Error: pg_wrapper: pg_receivewal was not found in
+        #   /usr/lib/postgresql/9.6/bin
+        # 1
+        # $ pg_receivexlog --version; echo $?
+        # pg_receivexlog (PostgreSQL) 9.6.3
+        # 0
+        # ```
+        #
+        # That means we should not only ensure the existence of the command,
+        # but we also need to invoke the command to see if it is a shim
+        # or not.
+
+        # Get the system path if needed
+        if path is None:
+            path = os.getenv('PATH')
+        # If the path is None at this point we have nothing to search
+        if path is None:
+            path = ''
+
+        # Search the requested executable in every directory present
+        # in path and return a Command object first occurrence that exists,
+        # is executable and runs without errors.
+        for path_entry in path.split(os.path.pathsep):
+            for cmd in cls.COMMAND_ALTERNATIVES:
+                full_path = barman.utils.which(cmd, path_entry)
+
+                # It doesn't exist try another
+                if not full_path:
+                    continue
+
+                # It exists, let's try invoking it with `--version` to check if
+                # it's real or not.
+                try:
+                    command = Command(full_path, path=path, check=True)
+                    command("--version")
+                    return command
+                except CommandFailedException:
+                    # It's only a inactive shim
+                    continue
+
+        # We don't have such a command
+        raise CommandFailedException(
+            'command not in PATH, tried: %s' %
+            ' '.join(cls.COMMAND_ALTERNATIVES))
+
     @classmethod
     def get_version_info(cls, path=None):
         """
@@ -726,28 +806,41 @@ class PostgreSQLClient(Command):
 
         :param str path: the PATH env
         """
-        if cls.COMMAND is None:
+        if cls.COMMAND_ALTERNATIVES is None:
             raise NotImplementedError(
                 "get_version_info cannot be invoked on %s" % cls.__name__)
 
-        version_info = dict.fromkeys(('full_path',
-                                      'full_version',
-                                      'major_version'),
-                                     None)
+        version_info = dict.fromkeys(
+            ('full_path', 'full_version', 'major_version'),
+            None)
 
         # Get the version string
         try:
-            command = Command(cls.COMMAND, path=path, check=True)
-            version_info['full_path'] = command.cmd
-            command("--version")
+            command = cls.find_command(path)
         except CommandFailedException as e:
-            _logger.debug("Error invoking %s: %s", cls.COMMAND, e)
+            _logger.debug("Error invoking %s: %s", cls.__name__, e)
             return version_info
 
+        version_info['full_path'] = command.cmd
         # Parse the full text version
-        full_version = command.out.strip().split()[-1]
-        version_info['full_version'] = Version(full_version)
+        try:
+            full_version = command.out.strip()
+            # Remove values inside parenthesis, they
+            # carries additional information we do not need.
+            full_version = re.sub(r'\s*\([^)]*\)', '', full_version)
+            full_version = full_version.split()[1]
+        except IndexError:
+            _logger.debug("Error parsing %s version output",
+                          version_info['full_path'])
+            return version_info
+
+        if not re.match(r'(\d+)(\.(\d+)|devel|beta|alpha|rc).*', full_version):
+            _logger.debug("Error parsing %s version output",
+                          version_info['full_path'])
+            return version_info
+
         # Extract the major version
+        version_info['full_version'] = Version(full_version)
         version_info['major_version'] = Version(barman.utils.simplify_version(
             full_version))
 
@@ -759,12 +852,12 @@ class PgBaseBackup(PostgreSQLClient):
     Wrapper class for the pg_basebackup system command
     """
 
-    COMMAND = 'pg_basebackup'
+    COMMAND_ALTERNATIVES = ['pg_basebackup']
 
     def __init__(self,
                  connection,
                  destination,
-                 command=COMMAND,
+                 command,
                  version=None,
                  app_name=None,
                  bwlimit=None,
@@ -798,6 +891,15 @@ class PgBaseBackup(PostgreSQLClient):
         # Set the backup destination
         self.args += ['-v', '--no-password', '--pgdata=%s' % destination]
 
+        if version and version >= Version("10"):
+            # If version of the client is >= 10 it would use
+            # a temporary replication slot by default to keep WALs.
+            # We don't need it because Barman already stores the full
+            # WAL stream, so we disable this feature to avoid wasting one slot.
+            self.args += ['--no-slot']
+            # We also need to specify that we do not want to fetch any WAL file
+            self.args += ['--wal-method=none']
+
         # The tablespace mapping option is repeated once for each tablespace
         if tbs_mapping:
             for (tbs_source, tbs_destination) in tbs_mapping.items():
@@ -822,12 +924,12 @@ class PgReceiveXlog(PostgreSQLClient):
     Wrapper class for pg_receivexlog
     """
 
-    COMMAND = "pg_receivexlog"
+    COMMAND_ALTERNATIVES = ["pg_receivewal", "pg_receivexlog"]
 
     def __init__(self,
                  connection,
                  destination,
-                 command=COMMAND,
+                 command,
                  version=None,
                  app_name=None,
                  synchronous=False,
@@ -881,7 +983,7 @@ class BarmanSubProcess(object):
     """
 
     def __init__(self, command=sys.argv[0], subcommand=None,
-                 config=None, args=None):
+                 config=None, args=None, keep_descriptors=False):
         """
         Build a specific wrapper for all the barman sub-commands,
         providing an unified interface.
@@ -891,6 +993,8 @@ class BarmanSubProcess(object):
         :param str config: path to the barman configuration file.
         :param list[str] args: a list containing the sub-command args
             like the target server name
+        :param bool keep_descriptors: whether to keep the subprocess stdin,
+            stdout, stderr descriptors attached. Defaults to False
         """
         # The config argument is needed when the user explicitly
         # passes a configuration file, as the child process
@@ -907,6 +1011,7 @@ class BarmanSubProcess(object):
         # * set it quiet with -q
         self.command = [sys.executable, command,
                         '-c', config, '-q', subcommand]
+        self.keep_descriptors = keep_descriptors
         # Handle args for the sub-command (like the server name)
         if args:
             self.command += args
@@ -918,10 +1023,18 @@ class BarmanSubProcess(object):
         _logger.debug("BarmanSubProcess: %r", self.command)
         # Redirect all descriptors to /dev/null
         devnull = open(os.devnull, 'a+')
+
+        additional_arguments = {}
+        if not self.keep_descriptors:
+            additional_arguments = {
+                'stdout': devnull,
+                'stderr': devnull
+            }
+
         proc = subprocess.Popen(
             self.command,
             preexec_fn=os.setsid, close_fds=True,
-            stdin=devnull, stdout=devnull, stderr=devnull)
+            stdin=devnull, **additional_arguments)
         _logger.debug("BarmanSubProcess: subprocess started. pid: %s",
                       proc.pid)
 
