@@ -1,4 +1,4 @@
-# Copyright (C) 2011-2017 2ndQuadrant Limited
+# Copyright (C) 2011-2020 2ndQuadrant Limited
 #
 # This file is part of Barman.
 #
@@ -33,7 +33,6 @@ import tempfile
 from functools import partial
 from multiprocessing import Lock, Pool
 
-import dateutil.parser
 import dateutil.tz
 
 from barman.command_wrappers import RsyncPgData
@@ -77,7 +76,7 @@ def _run_worker(job):
     # This is the entrypoint of the worker process. Since the KeyboardInterrupt
     # exceptions is handled by the main process, let's forget about Ctrl-C
     # here.
-    # When the parent process will receive a KeyboardInterrupt, is will ask
+    # When the parent process will receive a KeyboardInterrupt, it will ask
     # the pool to terminate its workers and then terminate itself.
     signal.signal(signal.SIGINT, signal.SIG_IGN)
     return _worker_callable(job)
@@ -87,10 +86,10 @@ class _RsyncJob(object):
     """
     A job to be executed by a worker Process
     """
-    def __init__(self, item, description,
+    def __init__(self, item_idx, description,
                  id=None, file_list=None, checksum=None):
         """
-        :param _RsyncCopyItem item: The copy item containing this job
+        :param int item_idx: The index of copy item containing this job
         :param str description: The description of the job, used for logging
         :param int id: Job ID (as in bucket)
         :param list[RsyncCopyController._FileItem] file_list: Path to the file
@@ -98,7 +97,7 @@ class _RsyncJob(object):
         :param bool checksum: Whether to force the checksum verification
         """
         self.id = id
-        self.item = item
+        self.item_idx = item_idx
         self.description = description
         self.file_list = file_list
         self.checksum = checksum
@@ -182,6 +181,10 @@ class _RsyncCopyItem(object):
         self.analysis_start_time = None
         self.analysis_end_time = None
 
+        # Ensure that the user specified the item class, since it is mandatory
+        # to correctly handle the item
+        assert self.item_class
+
     def __str__(self):
         # Prepare strings for messages
         formatted_class = self.item_class
@@ -192,8 +195,8 @@ class _RsyncCopyItem(object):
         formatted_class += ' directory' if self.is_directory else ' file'
 
         # Log the operation that is being executed
-        if self.item_class in(RsyncCopyController.PGDATA_CLASS,
-                              RsyncCopyController.PGCONTROL_CLASS):
+        if self.item_class in (RsyncCopyController.PGDATA_CLASS,
+                               RsyncCopyController.PGCONTROL_CLASS):
             return "%s: %s" % (
                 formatted_class, formatted_name)
         else:
@@ -215,9 +218,7 @@ class RsyncCopyController(object):
     # This regular expression is used to parse each line of the output
     # of a "rsync --list-only" call. This regexp has been tested with any known
     # version of upstream rsync that is supported (>= 3.0.4)
-    LIST_ONLY_RE = re.compile("""
-        (?x) # Enable verbose mode
-
+    LIST_ONLY_RE = re.compile(r"""
         ^ # start of the line
 
         # capture the mode (es. "-rw-------")
@@ -243,15 +244,13 @@ class RsyncCopyController(object):
         (?P<path>.+)
 
         $ # end of the line
-    """)
+    """, re.VERBOSE)
 
     # This regular expression is used to ignore error messages regarding
     # vanished files that are not really an error. It is used because
     # in some cases rsync reports it with exit code 23 which could also mean
     # a fatal error
-    VANISHED_RE = re.compile("""
-        (?x) # Enable verbose mode
-
+    VANISHED_RE = re.compile(r"""
         ^ # start of the line
         (
         # files which vanished before rsync start
@@ -265,10 +264,10 @@ class RsyncCopyController(object):
         |
         # final summary
         rsync\ error:\ .* \(code\ 23\)\ at\ main\.c\(\d+\)
-            \ \[generator=[^\]]+\]
+            \ \[(generator|receiver)=[^\]]+\]
         )
         $ # end of the line
-    """)
+    """, re.VERBOSE + re.IGNORECASE)
 
     def __init__(self, path=None, ssh_command=None, ssh_options=None,
                  network_compression=False,
@@ -305,6 +304,9 @@ class RsyncCopyController(object):
         self.retry_sleep = retry_sleep
         self.workers = workers
 
+        # Assume we are running with a recent rsync (>= 3.1)
+        self.rsync_has_ignore_missing_args = True
+
         self.item_list = []
         """List of items to be copied"""
 
@@ -385,7 +387,6 @@ class RsyncCopyController(object):
             and logging.
         :param str src: source directory.
         :param str dst: destination directory.
-        :param bwlimit: bandwidth limit to be enforced. (KiB)
         :param str item_class: If specified carries a meta information about
             what the object to be copied is.
         :param bool optional: Whether a failure copying this object should be
@@ -421,6 +422,14 @@ class RsyncCopyController(object):
         else:
             exclude = self.exclude or item.exclude
 
+        # Using `--ignore-missing-args` could fail in case
+        # the local or the remote rsync is older than 3.1.
+        # In that case we expect that during the analyze phase
+        # we get an error. The analyze code must catch that error
+        # and retry after flushing the rsync cache.
+        if self.rsync_has_ignore_missing_args:
+            args.append('--ignore-missing-args')
+
         # TODO: remove debug output or use it to progress tracking
         # By adding a double '--itemize-changes' option, the rsync
         # output will contain the full list of files that have been
@@ -446,6 +455,16 @@ class RsyncCopyController(object):
         self.rsync_cache[item] = rsync
         return rsync
 
+    def _rsync_set_pre_31_mode(self):
+        """
+        Stop using `--ignore-missing-args` and restore rsync < 3.1
+        compatibility
+        """
+        _logger.info("Detected rsync version less than 3.1. "
+                     "top using '--ignore-missing-args' argument.")
+        self.rsync_has_ignore_missing_args = False
+        self.rsync_cache.clear()
+
     def copy(self):
         """
         Execute the actual copy
@@ -456,7 +475,9 @@ class RsyncCopyController(object):
         # Create a temporary directory to hold the file lists.
         self.temp_dir = tempfile.mkdtemp(suffix='', prefix='barman-')
         # The following try block is to make sure the temporary directory
-        # will be removed on exit.
+        # will be removed on exit and all the pool workers
+        # have been terminated.
+        pool = None
         try:
             # Initialize the counters used by progress reporting
             self._progress_init()
@@ -510,23 +531,29 @@ class RsyncCopyController(object):
                 self.jobs_done.append(job)
 
         except KeyboardInterrupt:
-            # The parent process has been interrupted with a Ctrl-C. Since we
-            # handle this signal only from the main process (the workers will
-            # discard it), let's ask the pool to terminate its workers, wait
-            # for them to be really terminated, and than recover from the
-            # interruption in the usual way.
-            pool.terminate()
-            pool.join()
+            _logger.info("Copy interrupted by the user (safe before %s)",
+                         self.safe_horizon)
             raise
-        except:
+        except BaseException:
             _logger.info("Copy failed (safe before %s)", self.safe_horizon)
             raise
         else:
             _logger.info("Copy finished (safe before %s)", self.safe_horizon)
         finally:
-            # Clean tmp dir and log, exception management is delegated to
-            # the executor class
-            shutil.rmtree(self.temp_dir)
+            # The parent process may have finished naturally or have been
+            # interrupted by an exception (i.e. due to a copy error or
+            # the user pressing Ctrl-C).
+            # At this point we must make sure that all the workers have been
+            # correctly terminated before continuing.
+            if pool:
+                pool.terminate()
+                pool.join()
+            # Clean up the temp dir, any exception raised here is logged
+            # and discarded to not clobber an eventual exception being handled.
+            try:
+                shutil.rmtree(self.temp_dir)
+            except EnvironmentError as e:
+                _logger.error("Error cleaning up '%s' (%s)", self.temp_dir, e)
             self.temp_dir = None
 
             # Store the end time
@@ -542,7 +569,7 @@ class RsyncCopyController(object):
             which have one of the specified classes.
         :rtype: iter[_RsyncJob]
         """
-        for item in self.item_list:
+        for item_idx, item in enumerate(self.item_list):
 
             # Skip items of classes which are not required
             if include_classes and item.item_class not in include_classes:
@@ -561,7 +588,7 @@ class RsyncCopyController(object):
                 for i, bucket in enumerate(
                         self._fill_buckets(item.safe_list)):
                     phase_skipped = False
-                    yield _RsyncJob(item,
+                    yield _RsyncJob(item_idx,
                                     id=i,
                                     description=msg,
                                     file_list=bucket,
@@ -576,7 +603,7 @@ class RsyncCopyController(object):
                 for i, bucket in enumerate(
                         self._fill_buckets(item.check_list)):
                     phase_skipped = False
-                    yield _RsyncJob(item,
+                    yield _RsyncJob(item_idx,
                                     id=i,
                                     description=msg,
                                     file_list=bucket,
@@ -587,7 +614,7 @@ class RsyncCopyController(object):
             else:
                 # Copy the file using plain rsync
                 msg = self._progress_message("[%%s] %%s copy %s" % item)
-                yield _RsyncJob(item, description=msg)
+                yield _RsyncJob(item_idx, description=msg)
 
     def _fill_buckets(self, file_list):
         """
@@ -638,7 +665,7 @@ class RsyncCopyController(object):
 
         :type job: _RsyncJob
         """
-        item = job.item
+        item = self.item_list[job.item_idx]
         if job.id is not None:
             bucket = 'bucket %s' % job.id
         else:
@@ -722,7 +749,7 @@ class RsyncCopyController(object):
         """
         self.current_step += 1
         return "Copy step %s of %s: %s" % (
-                     self.current_step, self.total_steps, msg)
+            self.current_step, self.total_steps, msg)
 
     def _reuse_args(self, reuse_directory):
         """
@@ -782,9 +809,6 @@ class RsyncCopyController(object):
         :param _RsyncCopyItem item: information about a copy operation
         """
 
-        # Build the rsync object required for the analysis
-        rsync = self._rsync_factory(item)
-
         # If reference is not set we use dst as reference path
         ref = item.reuse
         if ref is None:
@@ -800,7 +824,7 @@ class RsyncCopyController(object):
         try:
             ref_hash = dict((
                 (item.path, item)
-                for item in self._list_files(rsync, ref)
+                for item in self._list_files(item, ref)
                 if item.mode[0] != 'd'))
         except (CommandFailedException, RsyncListFilesFailure) as e:
             # Here we set ref_hash to None, thus disable the code that marks as
@@ -831,7 +855,7 @@ class RsyncCopyController(object):
         # The `check_list` will contain all items that need
         # to be copied with checksum option enabled
         item.check_list = []
-        for entry in self._list_files(rsync, item.src):
+        for entry in self._list_files(item, item.src):
             # If item is a directory, we only need to save it in 'dir.list'
             if entry.mode[0] == 'd':
                 dir_list.write(entry.path + '\n')
@@ -839,8 +863,8 @@ class RsyncCopyController(object):
 
             # Add every file in the source path to the list of files
             # to be protected from deletion ('exclude_and_protect.filter')
-            exclude_and_protect_filter.write('P ' + entry.path + '\n')
-            exclude_and_protect_filter.write('- ' + entry.path + '\n')
+            exclude_and_protect_filter.write('P /' + entry.path + '\n')
+            exclude_and_protect_filter.write('- /' + entry.path + '\n')
 
             # If source item is older than safe_horizon,
             # add it to 'safe.list'
@@ -859,8 +883,13 @@ class RsyncCopyController(object):
             # destination, rsync will discover the difference in any case.
             # It is then safe to skip checksum check here.
             dst_item = ref_hash.get(entry.path, None)
-            if (dst_item is None or dst_item.size != entry.size or
-                    dst_item.date != entry.date):
+            if dst_item is None:
+                item.safe_list.append(entry)
+                continue
+
+            different_size = dst_item.size != entry.size
+            different_date = dst_item.date != entry.date
+            if different_size or different_date:
                 item.safe_list.append(entry)
                 continue
 
@@ -912,22 +941,43 @@ class RsyncCopyController(object):
             args.append('--checksum')
         self._rsync_ignore_vanished_files(rsync, src, dst, *args, check=True)
 
-    def _list_files(self, rsync, path):
+    def _list_files(self, item, path):
         """
         This method recursively retrieves a list of files contained in a
         directory, either local or remote (if starts with ':')
 
-        :param Rsync rsync: the Rsync object used to retrieve the list
+        :param _RsyncCopyItem item: information about a copy operation
         :param str path: the path we want to inspect
         :except CommandFailedException: if rsync call fails
         :except RsyncListFilesFailure: if rsync output can't be parsed
         """
         _logger.debug("list_files: %r", path)
-        # Use the --no-human-readable option to avoid digit groupings
-        # in "size" field with rsync >= 3.1.0.
-        # Ref: http://ftp.samba.org/pub/rsync/src/rsync-3.1.0-NEWS
-        rsync.get_output('--no-human-readable', '--list-only', '-r', path,
-                         check=True)
+
+        # Build the rsync object required for the analysis
+        rsync = self._rsync_factory(item)
+
+        try:
+            # Use the --no-human-readable option to avoid digit groupings
+            # in "size" field with rsync >= 3.1.0.
+            # Ref: http://ftp.samba.org/pub/rsync/src/rsync-3.1.0-NEWS
+            rsync.get_output('--no-human-readable', '--list-only',
+                             '-r', path, check=True)
+        except CommandFailedException:
+            # This could fail due to the local or the remote rsync
+            # older than 3.1. IF so, fallback to pre 3.1 mode
+            if self.rsync_has_ignore_missing_args and rsync.ret in (
+                    12,  # Error in rsync protocol data stream (remote)
+                    1):  # Syntax or usage error (local)
+                self._rsync_set_pre_31_mode()
+                # Recursive call, uses the compatibility mode
+                for item in self._list_files(item, path):
+                    yield item
+                return
+            else:
+                raise
+
+        # Cache tzlocal object we need to build dates
+        tzinfo = dateutil.tz.tzlocal()
         for line in rsync.out.splitlines():
             line = line.rstrip()
             match = self.LIST_ONLY_RE.match(line)
@@ -936,8 +986,18 @@ class RsyncCopyController(object):
                 # no exceptions here: the regexp forces 'size' to be an integer
                 size = int(match.group('size'))
                 try:
-                    date = dateutil.parser.parse(match.group('date'))
-                    date = date.replace(tzinfo=dateutil.tz.tzlocal())
+                    date_str = match.group('date')
+                    # The date format has been validated by LIST_ONLY_RE.
+                    # Use "2014/06/05 18:00:00" format if the sending rsync
+                    # is compiled with HAVE_STRFTIME, otherwise use
+                    # "Thu Jun  5 18:00:00 2014" format
+                    if date_str[0].isdigit():
+                        date = datetime.datetime.strptime(
+                            date_str, "%Y/%m/%d %H:%M:%S")
+                    else:
+                        date = datetime.datetime.strptime(
+                            date_str, "%a %b %d %H:%M:%S %Y")
+                    date = date.replace(tzinfo=tzinfo)
                 except (TypeError, ValueError):
                     # This should not happen, due to the regexp
                     msg = ("Unable to parse rsync --list-only output line "
@@ -1009,15 +1069,18 @@ class RsyncCopyController(object):
                 continue
             # Build a human readable name to refer to an item in the output
             ident = item.label
-            if (analysis_start is None or
-                    analysis_start > item.analysis_start_time):
+            if not analysis_start:
                 analysis_start = item.analysis_start_time
-            if (analysis_end is None or
-                    analysis_end < item.analysis_end_time):
+            elif analysis_start > item.analysis_start_time:
+                analysis_start = item.analysis_start_time
+
+            if not analysis_end:
                 analysis_end = item.analysis_end_time
+            elif analysis_end < item.analysis_end_time:
+                analysis_end = item.analysis_end_time
+
             stat['analysis_time_per_item'][ident] = total_seconds(
-                item.analysis_end_time -
-                item.analysis_start_time)
+                item.analysis_end_time - item.analysis_start_time)
         stat['analysis_time'] = total_seconds(analysis_end - analysis_start)
 
         # Calculate the time spent per job
@@ -1029,7 +1092,7 @@ class RsyncCopyController(object):
             # contained in self.item_list, as it has gone through two
             # pickling/unpickling cycle
             # Build a human readable name to refer to an item in the output
-            ident = job.item.label
+            ident = self.item_list[job.item_idx].label
             # If this is the first time we see this item we just store the
             # values from the job
             if ident not in item_data:
