@@ -1,4 +1,4 @@
-# Copyright (C) 2011-2017 2ndQuadrant Limited
+# Copyright (C) 2011-2020 2ndQuadrant Limited
 #
 # This file is part of Barman.
 #
@@ -19,9 +19,9 @@
 This module represents the barman diagnostic tool.
 """
 
+import datetime
 import json
 import logging
-import sys
 
 import barman
 from barman import fs, output
@@ -54,6 +54,7 @@ def exec_diagnose(servers, errors_list):
     except CommandFailedException as e:
         diagnosis['global']['system_info'] = {'error': repr(e)}
     diagnosis['global']['system_info']['barman_ver'] = barman.__version__
+    diagnosis['global']['system_info']['timestamp'] = datetime.datetime.now()
     # per server section
     for name in sorted(servers):
         server = servers[name]
@@ -75,7 +76,7 @@ def exec_diagnose(servers, errors_list):
                     command.get_system_info())
             except FsOperationFailed:
                 pass
-        # barman statuts information for the server
+        # barman status information for the server
         diagnosis['servers'][name]['status'] = server.get_remote_status()
         # backup list
         backups = server.get_available_backups(BackupInfo.STATUS_ALL)
@@ -87,5 +88,5 @@ def exec_diagnose(servers, errors_list):
         }
         # Release any PostgreSQL resource
         server.close()
-    output.info(json.dumps(diagnosis, sys.stdout, cls=BarmanEncoder, indent=4,
+    output.info(json.dumps(diagnosis, cls=BarmanEncoder, indent=4,
                            sort_keys=True))
